#pragma once

#include "Allocator.h"
#include <stdio.h>

template<typename T>
class Vector
{
public:
    Vector() = default;

    Vector(std::size_t size)
    {
        resize(size);
        m_size = size;
    }

    Vector(const Vector& v)
    {
        *this = v;
    }

    ~Vector()
    {
        clear();
    }

    T* begin() const
    {
        return m_data;
    }

    T* end() const
    {
        return m_data + m_size;
    }

    const T* data() const 
    {
        return m_data;
    }

    T* data() 
    {
        return m_data;
    }

    bool slice(std::size_t slice_idx, Vector& v) const
    {
        if (slice_idx >= m_size)
        {
            return false;
        }
        v.resize(slice_idx);
        for(std::size_t idx=0; idx<slice_idx;idx++)
        {
            v.m_data[idx] = m_data[idx];
        }
        v.m_size = slice_idx;
        return true;
    }

    bool slice(std::size_t slice_start, std::size_t slice_end, Vector& v) const
    {
        if (slice_end >= m_size)
        {
            return false;
        }
        v.resize(slice_end - slice_start);
        for(std::size_t idx=slice_start; idx<slice_end;idx++)
        {
            v.m_data[idx] = m_data[idx];
        }
        v.m_size = slice_end;
        return true;
    }

    std::size_t size() const 
    {
        return m_size;
    }

    std::size_t capacity() const 
    {
        return m_capacity;
    }

    bool empty() const
    {
        return m_size == 0;
    }

    void clear()
    {
        if (has_allocated())
        {
            m_allocator.delete_array(m_data);
            m_data = nullptr;
            m_capacity = 0;
            m_size = 0;
        }     
    }

    void resize(std::size_t size)
    {
        resize_capacity(size);
    }

    void resize(std::size_t size, const T& value)
    {
        resize_capacity(size);
        fill_with(value);
    }

    void extend(const Vector& other)
    {
        resize(m_size + other.m_size);
        for(std::size_t idx=0;idx<other.m_size;idx++)
        {
            m_data[idx + m_size] = other[idx];
        }
        m_size = m_size + other.m_size;
    }

    T pop_back()
    {
        const auto last = m_data[size() - 1];
        m_data[size() - 1] = T();
        m_size--;
        return last;
    }

    void push_back(const T& item)
    {
        if (!has_allocated())
        {
            resize_capacity(10);
        }
        else if (m_size >= m_capacity)
        {
            const std::size_t new_size = 1.5*m_size;
            resize_capacity(new_size);
        }
        m_data[m_size] = item;
        m_size++;      
    }

    const T& operator[] (const std::size_t idx) const
    {
        return m_data[idx];
    }

    T& operator[] (const std::size_t idx)
    {
        return m_data[idx];
    }

    Vector<T>& operator=(const Vector<T>& v)
    {
        resize(v.size());
        for(std::size_t idx=0; idx<v.size(); idx++)
        {
            m_data[idx] = v.m_data[idx];
        }
        m_size = v.size();
        return *this;
    }

    bool operator ==(const Vector<T>& other) const
    {
        if (m_size != other.m_size)
        {
            return false;
        }
        for(std::size_t idx=0; idx<m_size; idx++)
        {
            if (m_data[idx] != other[idx])
            {
                return false;
            }
        }
        return true;
    }

private:
    void fill_with(const T& value)
    {
        for(std::size_t idx = 0; idx<m_size; idx++)
        {
            m_data[idx] = value;
        }
    }

    bool has_allocated() const
    {
        return m_capacity > 0;
    }

    void resize_capacity(std::size_t new_capacity)
    {
        if (!has_allocated())
        {
            m_data = m_allocator.alloc_array(new_capacity);
            m_capacity = new_capacity;
        }
        else if (new_capacity != m_capacity)
        {
            auto temp = m_allocator.alloc_array(new_capacity);
            for(std::size_t idx=0; idx<new_capacity; idx++)
            {
                temp[idx] = m_data[idx];
            }
            auto old_size = m_size;
            clear();
            m_data = temp;
            m_capacity = new_capacity;
            m_size = old_size;
            if (old_size > m_capacity)
            {
                m_size = m_capacity;
            }
        }
        else if(m_size != m_capacity)
        {
            m_size = m_capacity;
        }
    }

    T* m_data{nullptr};
    Allocator<T> m_allocator;
    std::size_t m_size{0};
    std::size_t m_capacity{0};
};