#include "File.h"

//#include "FileLogger.h"
#include "ByteUtils.h"
#include "Directory.h"
#include "Result.h"

#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

class FileImpl
{
public:
    void do_open(const FileSystemPath& path, File::AccessMode accessMode)
    {
        int flags{0};
        if (accessMode == File::AccessMode::Read)
        {
            flags |= O_RDONLY;
        }
        else
        {
            flags |= O_WRONLY;
            flags |= O_CREAT;
        }

        m_fd = ::open(path.as_string().raw(), flags);
        if (accessMode == File::AccessMode::Read)
        {
            m_open_for_read = true;
        }
        else
        {
            m_open_for_write = true;
        }
    }

    void do_close()
    {
        ::close(m_fd);
        m_open_for_read = false;
        m_open_for_write = false;
    }

    bool is_ok() const
    {
        return m_valid;
    }

    bool is_open_for_read() const
    {
        return m_open_for_read;
    }

    bool is_open_for_write() const 
    {
        return m_open_for_write;
    }

    Result<std::size_t> do_read(VecBytes& bytes)
    {
        errno = 0;
        const auto rc = ::read(m_fd, bytes.data(), bytes.capacity());
        if (rc < 0)
        {
            const auto last_err = errno;
            String msg(::strerror(last_err));
            return Result<std::size_t>(Error(msg));
        }
        return Result<std::size_t>(rc);
    }

    std::size_t do_write(const VecBytes& bytes)
    {
        return ::write(m_fd, bytes.data(), bytes.size());
    }

    std::size_t do_write(const Vector<char>& bytes, int size = -1)
    {
        if (size > -1)
        {
            return ::write(m_fd, bytes.data(), size);
        }
        else
        {
            return ::write(m_fd, bytes.data(), bytes.size());
        }
    }

    void update_size()
    {
        struct stat buf;
        ::fstat(m_fd, &buf);
        m_size = buf.st_size;
    }

    bool m_open_for_write{false};
    bool m_open_for_read{false};
    bool m_valid{false};
    std::size_t m_size{0};
    bool m_is_open{false};
    int m_fd{-1};
};

File::File(const FileSystemPath& path)
    : m_impl(Ptr<FileImpl>::create()),
      m_path(path)
{
}

File::~File()
{
    close();
}

String File::getExtension() const
{
    return m_path.extension();
}

bool File::readBinary(VecBytes& buffer)
{
    auto ok = open(AccessMode::Read, true);
    if (!ok)
    {
        return false;
    }

    m_impl->update_size();
    buffer.resize(m_impl->m_size);
    const auto result = m_impl->do_read(buffer);
    buffer.resize(result.value());
    return true;
}

String File::readText()
{
    VecBytes buffer;
    auto ok = open(AccessMode::Read, true);
    if (!ok)
    {
        return {};
    }

    m_impl->update_size();
    buffer.resize(m_impl->m_size);
    const auto result = m_impl->do_read(buffer);
    if (!result.ok())
    {
        printf("Got error: %s\n", result.error().msg());
        return {};
    }

    buffer.resize(result.value());
    return String(buffer);
}

String File::dumpBinary()
{
    open(AccessMode::Read);

    String sstr;
    sstr << "Count | Binary | Decimal | Hex | ASCII \n";
    
    VecBytes buffer;
    readBinary(buffer);

    unsigned count = 0;
    /*
    for(const auto byte : buffer)
    {
        const auto ascii_val = std::isalpha(byte) ? byte : '.';
        String hex_sstr;
        hex_sstr << std::setfill('0') << std::setw(2) << std::hex << static_cast<int>(byte);

        sstr << count << " | " << ByteUtils::toString(byte) << " | " << static_cast<int>(byte) << " | " << hex_sstr.str() << " | " << ascii_val << '\n';
        if (count % 10 == 0)
        {
            sstr << "\n";
        }
        count++;
    }
    */
    close();
    return sstr;
}

Optional<Byte> File::readNextByte()
{
    if (!open(AccessMode::Read))
    {
        return {};
    }

    if (m_impl->is_ok())
    {
        VecBytes buffer(1);
        m_impl->do_read(buffer);
        if (buffer[0] == EOF)
        {
            return {};
        }
        else
        {
            return buffer[0];
        }
    }
    else
    {
        return {};
    }
}

bool File::open(AccessMode accessMode, bool binary)
{
    if (m_path.is_absolute() && !m_path.parent_path().exists())
    {
        Directory::create(m_path.parent_path(), true);
    }

    m_impl->do_open(m_path, accessMode);
    return true;
}

void File::close()
{
    m_impl->do_close();
}

FileFormat::Format File::inferFormat() const
{
    //const auto extension = getExtension();
    //return FileFormat::inferFormat(extension);
    return {};
}

void File::writeText(const String& text)
{
    bool had_to_open{false};
    if (!m_impl->is_open_for_write())
    {
        had_to_open = true;
        m_impl->do_open(m_path, File::AccessMode::Write);
    }

    m_impl->do_write(text.data(), text.data().size() - 1);

    if (had_to_open)
    {
        m_impl->do_close();
    }
}

Vector<String> File::readLines()
{
    Vector<String> content;

    /*
    if (!pathExists())
    {
        return {};
    }

    if(!open(AccessMode::Read))
    {
        return {};
    }

    String str;
    while(std::getline(*mInHandle, str))
    {
        content.push_back(str);
    }

    close();
    */
    return content;
}

String File::read()
{
    /*
    if (!pathExists())
    {
        return {};
    }

    if(!open(AccessMode::Read))
    {
        return {};
    }

    String buffer;
    buffer << mInHandle->rdbuf();

    close();
    */
    String buffer;
    return buffer;
}

bool File::pathExists() const
{
    return m_path.exists();
}
