#include "FileSystemPath.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>

FileSystemPath::FileSystemPath(const String& path)
    : m_path(path)
{

}

const String& FileSystemPath::as_string() const
{
    return m_path;
}

Result<bool> FileSystemPath::is_empty() const
{
    if (!is_regular_file())
    {
        return {Error("Requested empty check but target is not regular file.")};
    } 

    const auto size_result = get_size();
    if (!size_result.ok())
    {
        return {size_result.error().msg()};
    }
    return {size_result.value() == 0};
}

Result<size_t> FileSystemPath::get_size() const
{
    if (!is_regular_file())
    {
        return {Error("Requested size but target is not regular file.")};
    }

    struct stat buf;
    ::stat(m_path.raw(), &buf);
    return buf.st_size;
}

FileSystemPath FileSystemPath::current_dir()
{
    errno = 0;
    const auto path_max = ::pathconf(".", _PC_PATH_MAX);
    std::size_t size{0};
    if (path_max == -1)
    {
        size = 1024;
    }
    else if(path_max > 10240)
    {
        size = 10240;
    }
    else 
    {
        size = path_max;
    }

    Vector<char> buffer(path_max);
    const auto ret = ::getcwd(buffer.data(), path_max);
    return FileSystemPath(String(buffer.data()));
}

bool FileSystemPath::is_regular_file() const 
{
    struct stat path_stat;
    ::stat(m_path.raw(), &path_stat);
    return S_ISREG(path_stat.st_mode);
}

bool FileSystemPath::is_directory() const
{
    struct stat path_stat;
    ::stat(m_path.raw(), &path_stat);
    return S_ISDIR(path_stat.st_mode);
}

bool FileSystemPath::is_absolute() const 
{
    return false;
}

FileSystemPath FileSystemPath::parent_path() const
{
    const auto split = m_path.rsplit('/');
    return FileSystemPath(split.first());
}

String FileSystemPath::extension() const 
{
    const auto split = m_path.rsplit('.');
    String result(".");
    result += split.second();
    return result;
}

bool FileSystemPath::exists() const 
{
    return ::access(m_path.raw(), F_OK) == 0;
}

FileSystemPath FileSystemPath::join(const String& entry) const
{
    auto new_path = *this;
    new_path.m_path+=m_delimiter;
    new_path.m_path+=entry;
    return new_path;
}